/*
** $Id: rfc2045.c,v 1.2 1999/01/11 04:36:15 mrsam Exp $
*/
#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<ctype.h>

#include	"rfc2045.h"
#include	"rfc822.h"

#ifndef	RFC2045CHARSET
#define	RFC2045CHARSET	"us-ascii"
#endif

extern void cleanup();

void rfc2045_enomem()
{
	perror("malloc");
	cleanup();
	exit(1);
}

/*
	New RFC2045 structure.
*/

struct rfc2045 *rfc2045_alloc()
{
struct rfc2045 *p=(struct rfc2045 *)malloc(sizeof(struct rfc2045));

	if (!p)	rfc2045_enomem();

	/* Initialize everything to nulls, except for one thing */

	memset(p, '\0', sizeof(*p));

	p->pindex=1;	/* Start with part #1 */
	p->workinheader=1;
	/* Most of the time, we're about to read a header */

	return (p);
}

/*
	Unallocate the RFC2045 structure.  Recursively unallocate
	all sub-structures.  Unallocate all associated buffers.
*/

void rfc2045_free(struct rfc2045 *p)
{
struct rfc2045 *q;

	for (q=p->firstpart; q; q=q->next)
		rfc2045_free(q);
	if (p->content_boundary)	free(p->content_boundary);
	if (p->charset)			free(p->charset);
	if (p->content_transfer_encoding) free(p->content_transfer_encoding);
	if (p->content_name) free(p->content_name);
	if (p->content_type)	free(p->content_type);
	if (p->mime_version)	free(p->mime_version);
	if (p->workbuf)		free(p->workbuf);
	if (p->content_disposition) free(p->content_disposition);
	if (p->content_disposition_name) free(p->content_disposition_name);
	if (p->content_disposition_filename)
			free (p->content_disposition_filename);
	free(p);
}

/*
	Generic dynamic buffer append.
*/

static void add_buf(
	char **bufptr,	/* Buffer */
	size_t *bufsize,	/* Buffer's maximum size */
	size_t *buflen,		/* Buffer's current size */

	const char *p, size_t len)	/* Append this data */
{
	if (len + *buflen > *bufsize)
	{
	size_t	newsize=len+*buflen+256;
	char	*p= *bufptr ? (char *)realloc(*bufptr, newsize):
				(char *)malloc(newsize);

		if (!p)	rfc2045_enomem();
		*bufptr=p;
		*bufsize=newsize;
	}

	memcpy(*bufptr + *buflen, p, len);
	*buflen += len;
}

/* Append to the work buffer */

static void add_workbuf(struct rfc2045 *h, const char *p, size_t len)
{
	add_buf( &h->workbuf, &h->workbufsize, &h->workbuflen, p, len);
}

/* Append one character to the work buffer */

static void add_workbufch(struct rfc2045 *h, int c)
{
char cc= (char)c;

	add_workbuf(h, &cc, 1);
}

/*
	Generic function to duplicate contents of a string.
	The destination string may already be previously allocated,
	so unallocate it.
*/

static void set_string(char **p,
	const char *q)
{
	if (*p)	free(*p);

	*p=0;
	if (!q)	return;

	if ((*p=(char *)malloc(strlen(q)+1)) == 0)
		rfc2045_enomem();

	strcpy(*p, q);
}

/* Update byte counts for this structure, and all the superstructures */

static void update_counts(struct rfc2045 *p, size_t newcnt, size_t newendcnt)
{
	while (p)
	{
		p->endpos = newcnt;
		p->endbody = newendcnt;
		p=p->parent;
	}
}

/*
	Main entry point for RFC2045 parsing.  External data is fed
	by repetitively calling rfc2045_parse().

	rfc2045_parse() breaks up input into lines, and calls doline()
	to process each line.
*/

static void doline(struct rfc2045 *);

void rfc2045_parse(struct rfc2045 *h, const char *buf, size_t s)
{
size_t	l;

	while (s)
	{
		for (l=0; l<s; l++)
			if (buf[l] == '\n')	break;
		if (buf[l] == '\n')
		{
			++l;
			add_workbuf(h, buf, l);
			doline(h);
			h->workbuflen=0;
		}
		else
			add_workbuf(h, buf, l);
		buf += l;
		s -= l;
	}

	/* If our buffer has grown too big, see if that's because we're
	** in a form-data section, which we can see if we're parsing
	** form post in a CGI server.
	*/

	if (h->workbuflen > 512)
	{
	struct	rfc2045 *p;

		for (p=h; p->lastpart && !p->lastpart->workclosed;
				p=p->lastpart)
			;
		if (!p->workinheader &&
			p->content_disposition &&
			strcmp(p->content_disposition, "form-data") == 0)
		{
		int	l=h->workbuflen;
		int	i;

			/* We do need to make sure that the final \r\n gets
			** stripped off, so don't gobble up everything if
			** the last character we see is a \r
			*/

			if (h->workbuf[l-1] == '\r')
				--l;

			update_counts(p, p->endpos+l, p->endpos+l);
			p->informdata=1;
			for (i=0; l<h->workbuflen; l++)
				h->workbuf[i++]=h->workbuf[l];
			h->workbuflen=i;
		}
	}
}

/*
	Append a new RFC2045 subpart.  Adds new RFC2045 structure to the
	end of the list of existing RFC2045 substructures.
*/

static struct rfc2045 *append_part_noinherit(struct rfc2045 *p, size_t startpos){
struct rfc2045 *newp;

	newp=rfc2045_alloc();
	if (p->lastpart)
	{
		p->lastpart->next=newp;
		newp->pindex=p->lastpart->pindex+1;
	}
	else
	{
		p->firstpart=newp;
		newp->pindex=0;
	}
	p->lastpart=newp;
	newp->parent=p;

	/* Initialize source pointers */
	newp->startpos=newp->endpos=newp->startbody=newp->endbody=startpos;
	return (newp);
}

static struct rfc2045 *append_part(struct rfc2045 *p, size_t startpos)
{
struct rfc2045 *newp=append_part_noinherit(p, startpos);

	/* Substructures inherit content transfer encoding and character set */

	set_string(&newp->content_transfer_encoding,
			p->content_transfer_encoding);
	set_string(&newp->charset, p->charset);
	return (newp);
}

/*
	doline() processes next line in the RFC2045 message.

	Drills down the list of all the multipart messages currently open,
	and checks if the line is a boundary line for the given multipart.
	In theory the boundary line, if there is one, should be the boundary
	line only for the inner multipart only, but, this takes into account
	broken MIME messages.
*/

static void do_header(struct rfc2045 *);

static void doline(struct rfc2045 *p)
{
size_t	cnt=p->workbuflen;
char *c=p->workbuf;
size_t	n=cnt-1;	/* Strip \n (we always get at least a \n here) */
struct rfc2045 *newp;

	if (n && c[n-1] == '\r')	/* Strip trailing \r */
		--n;
	c[n]='\0';

	/* Before the main drill down loop before, look ahead and see if we're
	** in a middle of a form-data section.  */

	for (newp=p; newp->lastpart &&
			!newp->lastpart->workclosed; newp=newp->lastpart)
	{
		if (newp->content_boundary == 0 || newp->workinheader)
			continue;

		if (newp->lastpart->informdata)
		{
			p=newp->lastpart;
			p->informdata=0;
			break;
		}
	}

	/* Drill down until we match a boundary, or until we've reached
	the last RFC2045 section that has been opened.
	*/

	while (p->lastpart
		&& !p->lastpart->workclosed)
		/* Leftover trash -- workclosed is set when the final
		** terminating boundary has been seen */

	{
	size_t l;

		/* content_boundary may be set before the entire header
		** has been seen, so continue drilling down in that case
		*/

		if (p->content_boundary == 0 || p->workinheader)
		{
			p=p->lastpart;
			continue;
		}

		l=strlen(p->content_boundary);

		if (c[0] == '-' && c[1] == '-' &&
			strncasecmp(p->content_boundary, c+2, l) == 0)
		{

		/* Ok, we've found a boundary */

			if (strcmp(c+2+l, "--") == 0)
			{
			/* Last boundary */

				p->lastpart->workclosed=1;
				update_counts(p, p->endpos+cnt, p->endpos+n);
				return;
			}

		/* Create new RFC2045 section */

			newp=append_part(p, p->endpos+cnt);
			update_counts(p, p->endpos+cnt, p->endpos+n);

			/* The new RFC2045 section is Mime 1.0 compliant */

			if ((newp->mime_version=(char *)malloc(4)) == 0)
				rfc2045_enomem();

			strcpy(newp->mime_version, "1.0");
			return;
		}
		p=p->lastpart;
	}

	/* Ok, we've found the RFC2045 section that we're working with.
	** No what?
	*/

	if (! p->workinheader)
	{
		/* Processing body, just update the counts. */

		update_counts(p, p->endpos+cnt, p->endpos+n);
		return;
	}

	/* In the header */

	if ( *c == '\0' )	/* End of header, body begins.  Parse header. */
	{
		do_header(p);	/* Clean up any left over header line */
		p->workinheader=0;

		/* Message body starts right here */

		p->startbody=p->endpos+cnt;
		update_counts(p, p->startbody, p->startbody);

		/* Discard content type and boundary if I don't understand
		** this MIME flavor.
		*/

		if (!p->mime_version || atoi(p->mime_version) != 1)
		{
			set_string(&p->content_type, 0);
			set_string(&p->content_boundary, 0);
			set_string(&p->content_name, 0);
			set_string(&p->content_boundary, 0);
			set_string(&p->charset, 0);
			set_string(&p->content_disposition, 0);
			set_string(&p->content_disposition_name, 0);
			set_string(&p->content_disposition_filename, 0);
		}

		/* Normally, if we don't have a content_type, default it
		** to text/plain.  However, if the multipart type is
		** multipart/digest, it is message/rfc822.
		*/

		if (p->mime_version && !p->content_type)
		{
		char	*q="text/plain";

			if (p->parent && p->parent->content_type &&
				strcmp(p->parent->content_type,
					"multipart/digest") == 0)
				q="message/rfc822";
			set_string(&p->content_type, q);
		}

		/* If this is not a multipart section, we don't want to
		** hear about any boundaries
		*/

		if (!p->content_type ||
			strncmp(p->content_type, "multipart/", 10))
			set_string(&p->content_boundary, 0);

		/* If this section's a message, we will expect to see
		** more RFC2045 stuff, so create a nested RFC2045 structure,
		** and indicate that we expect to see headers.
		*/

		if (p->content_type &&
			strncmp(p->content_type, "message/", 8) == 0)
		{
			newp=append_part_noinherit(p, p->startbody);
			newp->workinheader=1;
		}

		/*
		** If this is a multipart message (boundary defined),
		** create a RFC2045 structure for the pseudo-section
		** that precedes the first boundary line.
		*/

		else if (p->content_boundary)
		{
			newp=append_part(p, p->startbody);
			newp->workinheader=0;
			newp->isdummy=1;
					/* It's easier just to create it. */
		}
		return;
	}

	/* RFC822 header continues */

	update_counts(p, p->endpos + cnt, p->endpos+n);

	/* If this header line starts with a space, append one space
	** to the saved contents of the previous line, and append this
	** line to it.
	*/

	if (isspace((int)(unsigned char)*c))
	{
		add_buf(&p->header, &p->headersize, &p->headerlen, " ", 1);
	}
	else
	{
	/* Otherwise the previous header line is complete, so process it */

		do_header(p);
		p->headerlen=0;
	}

	/* Save this line in the header buffer, because the next line
	** could be a continuation.
	*/

	add_buf( &p->header, &p->headersize, &p->headerlen, c, n);
}

/***********************************************************************/

/*
** paste_tokens() - recombine an array of RFC822 tokens back as a string.
** (Comments) are ignored.
*/

static char *paste_tokens(struct rfc822 *h, int start, int cnt)
{
int	l;
int	i;
char	*p;

	/* Calculate string size */

	l=1;
	for (i=0; i<cnt; i++)
	{
		if (h->tokens[start+i].token == '(')
			continue;

		if (rfc822_is_atom(h->tokens[start+i].token))
			l += h->tokens[start+i].len;
		else
			l++;
	}

	/* Do it */

	p=( char *)malloc(l);
	if (!p)
	{
		/* The RFC822 structure we've allocated was temporary.
		** Error condition - free it, so we don't leak memory.
		*/

		rfc822_free(h);
		rfc2045_enomem();
	}
	l=0;

	for (i=0; i<cnt; i++)
	{
		if (h->tokens[start+i].token == '(')
			continue;

		if (rfc822_is_atom(h->tokens[start+i].token))
		{
		int l2=h->tokens[start+i].len;

			memcpy(p+l, h->tokens[start+i].ptr, l2);
			l += l2;
		}
		else	p[l++]=h->tokens[start+i].token;
	}
	p[l]=0;
	return (p);
}

/* Various permutations of the above, including forcing the string to
** lowercase
*/

static char *lower_paste_tokens(struct rfc822 *h, int start, int cnt)
{
char	*p=paste_tokens(h, start, cnt);
char	*q;

	for (q=p; *q; q++)
		*q=tolower(*q);
	return (p);
}

static char *paste_token(struct rfc822 *h, int i)
{
	if (i >= h->ntokens)	return (0);
	return (paste_tokens(h, i, 1));
}

static char *lower_paste_token(struct rfc822 *h, int i)
{
char *p=paste_token(h, i);
char *q;

	for (q=p; *q; q++)
		*q=tolower(*q);
	return (p);
}

/*
	doheader() - process completed RFC822 header.
*/

static void mime_version(struct rfc2045 *, struct rfc822 *);
static void content_type(struct rfc2045 *, struct rfc822 *);
static void content_transfer_encoding(struct rfc2045 *, struct rfc822 *);
static void content_disposition(struct rfc2045 *, struct rfc822 *);

static void do_header(struct rfc2045 *p)
{
struct rfc822 *header;
char	*t;

	if (p->headerlen == 0)	return;
	add_buf( &p->header, &p->headersize, &p->headerlen, "", 1);
				/* 0 terminate */

	/* Parse the header line according to RFC822 */

	header=rfc822_alloc(p->header, NULL);

	if (!header)	return;	/* Broken header */

	if (header->ntokens < 2 ||
		header->tokens[0].token ||
		header->tokens[1].token != ':')
	{
		rfc822_free(header);
		return;	/* Broken header */
	}

	t=lower_paste_token(header, 0);
	if (strcmp(t, "mime-version") == 0)
	{
		free(t);
		mime_version(p, header);
	}
	else if (strcmp(t, "content-type") == 0)
	{
		free(t);
		content_type(p, header);
	} else if (strcmp(t, "content-transfer-encoding") == 0)
	{
		free(t);
		content_transfer_encoding(p, header);
	} else if (strcmp(t, "content-disposition") == 0)
	{
		free(t);
		content_disposition(p, header);
	}
	else	free(t);
	rfc822_free(header);
}

/* Mime-Version: and Content-Transfer-Encoding: headers are easy */

static void mime_version(struct rfc2045 *p, struct rfc822 *header)
{
char	*vers=paste_tokens(header, 2, header->ntokens-2);

	if (p->mime_version)	free(p->mime_version);
	p->mime_version=vers;
}

static void content_transfer_encoding(struct rfc2045 *r,
				struct rfc822 *header)
{
char	*p;

	p=lower_paste_tokens(header, 2, header->ntokens-2);
	if (r->content_transfer_encoding)
		free(r->content_transfer_encoding);
	r->content_transfer_encoding=p;
}

/* Dig into the content_type header */

static void parse_content_header(struct rfc2045 *r, struct rfc822 *header,
	void (*init_token)(struct rfc2045 *, char *),
	void (*init_parameter)(struct rfc2045 *, const char *,
			struct rfc822 *, int, int))
{
int	start;
int	i, j;
char	*p;

	/* Look for the 1st ; */

	for (start=2; start < header->ntokens; start++)
		if (header->tokens[start].token == ';')
			break;

	/* Everything up to the 1st ; is the content type */

	p=lower_paste_tokens(header, 2, start-2);
	(*init_token)(r, p);
	if (start < header->ntokens) start++;

	/* Handle the remainder of the Content-Type: header */

	while (start < header->ntokens)
	{
		/* Look for next ; */

		for (i=start; i<header->ntokens; i++)
			if (header->tokens[i].token == ';')
				break;
		j=start;
		if (j < i)
		{
			++j;

			/* We only understand <atom>= */

			while (j < i && header->tokens[j].token == '(')
				++j;
			if (j < i && header->tokens[j].token == '=')
			{
				++j;
				p=lower_paste_token(header, start);
				(*init_parameter)(r, p, header, j, i-j);
				free(p);
			}
		}
		if ( i<header->ntokens ) ++i;	/* Skip over ; */
		start=i;
	}
}

/* Dig into the content_type header */

static void save_content_type(struct rfc2045 *, char *);
static void save_content_type_parameter( struct rfc2045 *, const char *,
	struct rfc822 *, int, int);

static void content_type(struct rfc2045 *r, struct rfc822 *header)
{
	parse_content_header(r, header, &save_content_type,
		&save_content_type_parameter);
}

static void save_content_type(struct rfc2045 *r, char *content_type)
{
	if (r->content_type)	free(r->content_type);
	r->content_type=content_type;
}

static void save_content_type_parameter(
	struct rfc2045 *r, const char *name,
	struct rfc822 *header, int start, int len)
{
char	*p;

	if (strcmp(name, "charset") == 0)
	{
		p=lower_paste_tokens(header, start, len);
		set_string(&r->charset, 0);
		r->charset=p;
		return;
	}

	if (strcmp(name, "boundary") == 0)
	{
		p=lower_paste_tokens(header, start, len);
		set_string(&r->content_boundary, 0);
		r->content_boundary=p;
		return;
	}

	if (strcmp(name, "name") == 0)
	{
		p=paste_tokens(header, start, len);
		set_string(&r->content_name, 0);
		r->content_name=p;
		return;
	}
}

/* Dig into content-disposition */

static void save_content_disposition(struct rfc2045 *, char *);
static void save_content_disposition_parameter( struct rfc2045 *, const char *,
	struct rfc822 *, int, int);

static void content_disposition(struct rfc2045 *r, struct rfc822 *header)
{
	parse_content_header(r, header, &save_content_disposition,
		&save_content_disposition_parameter);
}

static void save_content_disposition(struct rfc2045 *r,
					char *content_disposition)
{
	if (r->content_disposition)	free(r->content_disposition);
	r->content_disposition=content_disposition;
}

static void save_content_disposition_parameter(
	struct rfc2045 *r, const char *name,
	struct rfc822 *header, int start, int len)
{
char	*p;

	if (strcmp(name, "name") == 0)
	{
		p=paste_tokens(header, start, len);
		set_string(&r->content_disposition_name, 0);
		r->content_disposition_name=p;
		return;
	}

	if (strcmp(name, "filename") == 0)
	{
		p=paste_tokens(header, start, len);
		set_string(&r->content_disposition_filename, 0);
		r->content_disposition_filename=p;
		return;
	}
}

/* -------------------- */

#define	GETINFO(s, def) ( (s) && (*s) ? (s):def)

void rfc2045_mimeinfo(const struct rfc2045 *p,
	const char **content_type_s,
	const char **content_transfer_encoding_s,
	const char **charset_s)
{
	*content_type_s=GETINFO(p->content_type, "text/plain");
	*content_transfer_encoding_s=GETINFO(p->content_transfer_encoding,
						"8bit");
	*charset_s=GETINFO(p->charset, RFC2045CHARSET);
}

const char *rfc2045_defaultcharset()
{
	return (RFC2045CHARSET);
}

const char *rfc2045_boundary(const struct rfc2045 *p)
{
	return (GETINFO(p->content_boundary, ""));
}

void rfc2045_dispositioninfo(const struct rfc2045 *p,
	const char **disposition_s,
	const char **disposition_name_s,
	const char **disposition_filename_s)
{
	*disposition_s=p->content_disposition;
	*disposition_name_s=p->content_disposition_name;
	*disposition_filename_s=p->content_disposition_filename;
}

const char *rfc2045_contentname(const struct rfc2045 *p)
{
const	char *q=p->content_name;

	if (!q)	q="";
	return (q);
}

void rfc2045_mimepos(const struct rfc2045 *p,
	off_t *start_pos, off_t *end_pos, off_t *start_body)
{
	*start_pos=p->startpos;
	*end_pos=p->endpos;
	if (p->parent)	/* MIME parts do not have the trailing CRLF */
		*end_pos=p->endbody;
	*start_body=p->startbody;
}

unsigned rfc2045_mimepartcount(const struct rfc2045 *p)
{
const struct rfc2045 *q;
unsigned n=0;

	for (q=p->firstpart; q; q=q->next)	++n;
	return (n);
}

static void decode_raw(struct rfc2045 *p, const char *s, size_t l)
{
	if (s && l)	(*p->udecode_func)(s,l);
}

static const char xdigit[]="0123456789ABCDEF";

static void do_decode_qp(struct rfc2045 *p)
{
char	*a, *b, *c, *end;
int	d;

	end=p->workbuf + p->workbuflen;
	for (a=b=p->workbuf; a < end; )
	{
		if (*a != '=')
		{
			*b++ = *a++;
			continue;
		}
		++a;
		if (!*a || a >= end || isspace((int)(unsigned char)*a))
			break;

		if ((c=strchr(xdigit, *a)) == 0) continue;
		d= (c-xdigit)*16;
		++a;
		if (!*a || a >= end)
			break;
		if ((c=strchr(xdigit, *a)) == 0) continue;
		d += c-xdigit;
		++a;
		*b++=d;
	}
	p->workbuflen= b-p->workbuf;
	(*p->udecode_func)(p->workbuf, p->workbuflen);
	p->workbuflen=0;
}

static unsigned char decode64tab[256];
static int decode64tab_init=0;

/* When we have enough base64-encoded data in the buffer, decode it. */

static void do_decode_base64(struct rfc2045 *p)
{
size_t	i, j;
char	a,b,c;
size_t	k;

	if (!decode64tab_init)
	{
		for (i=0; i<256; i++)	decode64tab[i]=100;
		for (i=0; i<64; i++)
			decode64tab[ (int)
				("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"[i])				]=i;
		decode64tab[ (int)'=' ] = 99;
	}

	/* Remove everything except base64encoded data */

	for (i=j=0; i<p->workbuflen; i++)
		if (decode64tab[(int)(unsigned char)p->workbuf[i]] < 100)
			p->workbuf[j++]=p->workbuf[i];


	p->workbuflen=j;

	/* Decode the data, in 4-byte pieces */

	i=j / 4;
	i=i*4;
	k=0;
	for (j=0; j<i; j += 4)
	{
	int	w=decode64tab[(int)(unsigned char)p->workbuf[j]];
	int	x=decode64tab[(int)(unsigned char)p->workbuf[j+1]];
	int	y=decode64tab[(int)(unsigned char)p->workbuf[j+2]];
	int	z=decode64tab[(int)(unsigned char)p->workbuf[j+3]];

		a= (w << 2) | (x >> 4);
		b= (x << 4) | (y >> 2);
		c= (y << 6) | z;
		p->workbuf[k++]=a;
		if ( p->workbuf[j+2] != '=')
			p->workbuf[k++]=b;
		if ( p->workbuf[j+3] != '=')
			p->workbuf[k++]=c;
	}
	(*p->udecode_func)(p->workbuf, k);

	/* Anything left?  Move it to the start of the buffer */

	k=0;
	while (j < p->workbuflen)
		p->workbuf[k++]=p->workbuf[j++];
	p->workbuflen=k;
}

static void decode_qp(struct rfc2045 *p, const char *s, size_t l)
{
size_t	start,i;

	if (!s)
	{
		do_decode_qp(p);
		return;
	}

	for (start=0; start<l; )
	{
		for (i=start; i<l; i++)
		{
			if (s[i] != '\n') continue;

			add_workbuf(p, s+start, i-start);
			add_workbufch(p, '\n');
			do_decode_qp(p);
			start= ++i;
			break;
		}
		add_workbuf(p, s+start, i-start);
		start=i;
	}
}

static void decode_base64(struct rfc2045 *p, const char *s, size_t l)
{
	if (!s)
	{
		do_decode_base64(p);
		return;
	}

	add_workbuf(p, s, l);
	if (p->workbuflen > 256)
		do_decode_base64(p);
}

void rfc2045_cdecode_start(struct rfc2045 *p, void (*u)(const char *, size_t))
{
	p->udecode_func=u;
	p->decode_func= &decode_raw;
	p->workbuflen=0;
	if (p->content_transfer_encoding)
	{
		if (strcmp(p->content_transfer_encoding,
				"quoted-printable") == 0)
			p->decode_func= &decode_qp;
		else if (strcmp(p->content_transfer_encoding, "base64") == 0)
			p->decode_func= &decode_base64;
	}
}

void rfc2045_cdecode_end(struct rfc2045 *p)
{
	(*p->decode_func)(p, NULL, 0);
}

void rfc2045_cdecode(struct rfc2045 *p, const char *s, size_t l)
{
	if (s && l)	(*p->decode_func)(p, s, l);
}
